#Typing imports
import typing as typ

#External imports
import qdarkstyle                                 # type: ignore
import sqlparse                                   # type: ignore
import time
from PyQt5 import QtCore, QtGui, QtWidgets, QtSql # type: ignore
#Internal Imports
from catalog.datalog.db_utils import Query, realDB
from catalog.misc.sql_utils import *
from catalog.misc.atoms import new_view
from catalog.gui.gendata_form import GenDataForm
from catalog.gui.launch_form import LaunchForm

user = os.environ['USER']


class JobMakerGUI(QtWidgets.QWidget):
    def __init__(self
                ,query       : str         = """SELECT job_id,
                                                       job_name,
                                                       USER,
                                                       stordir
                                                FROM job
                                                JOIN relax_job USING (job_id)
                                                JOIN calc USING (calc_id)
                                                JOIN calc_other USING (calc_other_id)
                                                JOIN finaltraj USING (job_id)
                                                JOIN struct USING (struct_id)
                                                LEFT JOIN surface USING (struct_id)
                                                LEFT JOIN bulk USING (struct_id)
                                                LEFT JOIN molecule USING (struct_id)
                                                WHERE 1
                                                LIMIT 100"""
                ,connectinfo : ConnectInfo = realDB
                ,parent      : typ.Any     = None
                ) -> None:
        QtWidgets.QWidget.__init__(self, parent)
        self.setupUi()


        #Build the table
        self.connectinfo = connectinfo
        self._connect_to_db()
        self.connect_count = 1
        
        self.model = QtSql.QSqlQueryModel()
        start = time.time()
        self.expert_editor.appendPlainText(query)
        self.limit = 100
        self._update_query()
        self._beautify()
        self._build_table()
        self._connect_signals()

    def _connect_signals(self) -> None:
        self.push_add_constraint.clicked.connect(lambda x: self.create_scroll_area(self.n_constraints+1))
        self.push_remove_constraint.clicked.connect(lambda x: self.create_scroll_area(max([3,self.n_constraints-1])))
        self.table.clicked.connect(self._table_click)
        self.push_quit.clicked.connect(self._quit)
        self.push_view.clicked.connect(self._view)
        self.push_launch.clicked.connect(self._launch)
        self.push_query_expert.clicked.connect(self._update_query)
        self.push_query.clicked.connect(self._update_query)

        self.regex_edit.textChanged.connect(self.on_lineEdit_textChanged)
        self.regex_combobox.currentIndexChanged.connect(self.on_comboBox_currentIndexChanged)
        self.horizontalHeader = self.table.horizontalHeader()
        self.horizontalHeader.sectionClicked.connect(self.on_view_horizontalHeader_sectionClicked)

    def keyPressEvent(self, event : typ.Any)->None:
        k = event.key()
        m = int(event.modifiers())
        if QtGui.QKeySequence(m+k) == QtGui.QKeySequence('Ctrl+Return'):
            self._update_query()
        elif QtGui.QKeySequence(m+k) == QtGui.QKeySequence('Ctrl+B'):
            self._beautify()
        elif QtGui.QKeySequence(m+k) == QtGui.QKeySequence('Ctrl+W'):
            self._quit(event)
        elif QtGui.QKeySequence(m+k) == QtGui.QKeySequence('Ctrl+C'):
            self._copy()

    def _connect_to_db(self) -> None:
        db = QtSql.QSqlDatabase.addDatabase("QMYSQL")
        db.setDatabaseName(self.connectinfo.db)
        db.setHostName(self.connectinfo.host)
        db.setUserName(self.connectinfo.user)
        db.setPassword(self.connectinfo.passwd)
        db.open()

    def _build_table(self)->None:
        self.table.setWordWrap(True)
        self.proxy = QtCore.QSortFilterProxyModel(self)
        self.proxy.setSourceModel(self.model)
        self.table.setModel(self.proxy)
        self.table.setSortingEnabled(True)
        #Remove editing capabilities and set only row selection
        self.table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.table.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)

    def _table_click(self,event : typ.Any)->None:
        model_index = self.table.model().mapToSource(event)
        self.clicked_model_index = model_index

    def _beautify(self)->None:
        command        = self.expert_editor.toPlainText()
        command        = command.strip('\n')
        self.expert_editor.clear()
        if command == '':
            command = """
            SELECT job_id,
                   job_name,
                   USER,
                   stordir
            FROM job
            JOIN relax_job USING (job_id)
            JOIN calc USING (calc_id)
            JOIN calc_other USING (calc_other_id)
            JOIN finaltraj USING (job_id)
            JOIN struct USING (struct_id)
            LEFT JOIN surface USING (struct_id)
            LEFT JOIN bulk USING (struct_id)
            LEFT JOIN molecule USING (struct_id)
            WHERE USER = '{}'
            ORDER BY TIMESTAMP desc
            LIMIT 100
            """.format(user)
            command        = command.strip('\n')
        self.expert_editor.appendPlainText(sqlparse.format(command.strip('\n'), reindent=True, keyword_case='upper'))

    def _update_query(self)->None:
        command = self._get_sqlcommand()
        if not 'limit' in command:
            command += ' limit {}'.format(self.limit)
        if 'update' in command or 'drop' in command:
            self.set_message('Query Failed!\nOnly select statements allowed',clear = True)
            return
        start = time.time()
        self.model.setQuery(command)
        fetchtime = round(time.time()-start,3)
        #Print to Message box
        if self.model.lastError().isValid():
            if 'MySQL server has gone away QMYSQL' in self.model.lastError().text() and self.connect_count<5:
                self.connect_count += 1
                self._connect_to_db()
                return self._update_query()
            self.set_message('Query Failed\n',clear = True)
            self.set_message(sqlparse.format(command, reindent=True, keyword_case='upper'))
            self.set_message(self.model.lastError().text())
        else:
            self.set_message('Query Executed Succesfully\n{} Rows Returned\n{} second\n'.format(self.model.rowCount(),fetchtime),clear = True)
            self.set_message(sqlparse.format(command, reindent=True, keyword_case='upper'))
        self.table.resizeColumnsToContents()
        #Set the regex filter box
        self.regex_combobox.clear()
        if self.model.rowCount()>0:
            self.regex_combobox.addItems([self.model.record(0).fieldName(x) for x in range(self.model.record(0).count())])

    def _get_sqlcommand(self) -> str:
        if self.tabWidget.currentIndex() == 0:
            return self.expert_editor.toPlainText().lower()
        else:
            select_cmd = self.select_edit.text().lower()
            where_cmd = ''
            for n_con in range(self.n_constraints):
                col_name  = self.constraint_columns[n_con].text().replace(' ','')
                condition = self.constraint_conditions[n_con].currentText()
                if condition   == 'Equal':
                    condition = '='
                elif condition == 'Not Equal':
                    condition = '!='
                val       = self.constraint_edits[n_con].text().replace(' ','').replace('\'','')
                if all(map(lambda x: not x == '',[col_name,condition,val])):
                    if not where_cmd == '':
                        where_cmd += ' AND '
                    if '(' in val or '[' in val:
                        where_cmd  += " {} {} {} ".format(col_name,condition,val)
                    else:
                        where_cmd  += " {} {} '{}' ".format(col_name,condition,val)
            group_by_cmd = self.group_by_edit.text().replace(' ','')
            if where_cmd:
                where_cmd = ' WHERE '+where_cmd
            if group_by_cmd:
                group_by_cmd = ' GROUP BY '+group_by_cmd
            limit_cmd = self.limit_edit.text().lower()
            if limit_cmd:
                limit_cmd = ' LIMIT '+limit_cmd
            cmd = """
            SELECT {}
            FROM job
            JOIN relax_job USING (job_id)
            JOIN calc USING (calc_id)
            JOIN calc_other USING (calc_other_id)
            JOIN finaltraj USING (job_id)
            JOIN struct USING (struct_id)
            LEFT JOIN surface USING (struct_id)
            LEFT JOIN bulk USING (struct_id)
            LEFT JOIN molecule USING (struct_id)
            {}
            {}
            {}
             """.format(select_cmd,where_cmd,group_by_cmd,limit_cmd)
            return cmd
    def _copy(self)->None:
        sys_clip = QtWidgets.QApplication.clipboard()
        current_cell_selected = str(self.model.record(self.clicked_model_index.row()).value(self.clicked_model_index.column()))
        sys_clip.setText(current_cell_selected)

    def _get_selected_job_ids(self)->list:
        indexes          = self.table.selectedIndexes()
        if len(indexes)>0:
            selected_job_ids = set()                                                # type: ignore
            for index in sorted(indexes):
                model_index = self.table.model().mapToSource(index)
                record      = self.model.record(model_index.row())
                if record.contains('job_id'):
                    selected_job_ids.update([self.model.record(model_index.row()).value('job_id')])
                else:
                    self.set_message('!!!!WARNING!!!!\nPlease include the column job_id in \nyour query to launch or view jobs\n',clear =True)
                    return []
            selected_job_ids = list(selected_job_ids)                               # type: ignore
        else:
            selected_job_ids = []
        return selected_job_ids

    def _view(self,event : typ.Any)->None:
        selected_job_ids = self._get_selected_job_ids()
        if len(selected_job_ids)>0:
            temp_query       = Query(constraints = [JOB_IN_(selected_job_ids)])     # type: ignore
            atoms            = temp_query.make_atoms()
            new_view(atoms)
        else:
            self.set_message("No Jobs selected\n")


    def _launch(self, event : typ.Any)-> None:
            """
            Halts the program
            """
            selected_job_ids = self._get_selected_job_ids()
            if self.central_tabWidget.count() == 1:
                if len(selected_job_ids)>0:
                    #Open up a dialog box to ask the user which file to view
                    dialogbox         = QtWidgets.QInputDialog()
                    job_types = ['Adsorb'
                                ,'Bare'
                                ,'Bare All Facets'
                                ,'Vacancy'         ]
                    job_type, ok = dialogbox.getItem(self, 'Job Type Selector','Select a job type to submit:',job_types,0,False)
                    #View that traj file
                    if job_type and ok:
                        form = self.central_tabWidget.addTab(GenDataForm(job_type = job_type, job_ids = selected_job_ids,parent = self),"Launch Form")
                        self.central_tabWidget.setCurrentIndex(1)
                else:
                    self.set_message('Please select a job to relaunch')
            else:
                self.set_message('Please close the existing Form to launch new jobs')

    def _quit(self,
              event : typ.Any) -> None:
        """
        Halts the program
        """
        reply = self.check_with_user()
        if reply:
            self.close()
            QtWidgets.QApplication.exit()

    @staticmethod
    def check_with_user(message : str = 'Are you sure you want to quit?') -> bool:
        reply = QtWidgets.QMessageBox.question(None, 'Message',
                     message, QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.No)
        if reply == QtWidgets.QMessageBox.Yes:
            return True
        else:
            return False

    def set_message(self
                   ,text  : str  = ''
                   ,clear : bool = False
                   ) -> None:
        if clear:
            self.message_box.clear()
        self.message_box.appendPlainText(text)

    def setupUi(self) -> None:
        """
        Create the UI for the QueryGUI object
        """
        self.setObjectName("Query")
        self.resize(1900, 732)
        self.setWindowTitle( "Query GUI")

        #Set Fonts for Push Buttons and Titles
        self.header_font = QtGui.QFont()
        self.header_font.setFamily("Helvetica Neue")
        self.header_font.setPointSize(15)

        # self.centralwidget = QtWidgets.QWidget(self)
        # self.centralwidget.setFont(self.header_font)
        # self.centralwidget.setObjectName("centralwidget")
        self.central_layout = QtWidgets.QHBoxLayout(self)
        self.central_layout.setObjectName("central_layout")
        self.central_tabWidget = QtWidgets.QTabWidget()
        self.central_tabWidget.tabBar().hide()
        self.central_layout.addWidget(self.central_tabWidget)

        self.query_tab        = QtWidgets.QWidget()
        self.query_tab_layout = QtWidgets.QHBoxLayout(self.query_tab)

        ##################
        #Create Left Frame and set its layout
        ##################
        self.left_frame = QtWidgets.QFrame()
        self.left_frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        # self.left_frame.setFrameShape(QtWidgets.QFrame.Panel)
        # self.left_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        # self.left_frame.setLineWidth(2)
        # self.left_frame.setMidLineWidth(1)
        self.left_frame_layout = QtWidgets.QVBoxLayout(self.left_frame)

        #Create push button box and fill it with push buttons
        ##################
        self.push_group = QtWidgets.QGroupBox(self.left_frame)
        self.push_group.setFont(self.header_font)

        self.push_group.setObjectName("push_group")
        self.push_group_layout = QtWidgets.QVBoxLayout(self.push_group)
        self.push_group_layout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.push_group_layout.setObjectName("push_group_layout")

        #Create push_view button
        ##################
        self.push_view = QtWidgets.QPushButton(self.push_group)
        self.push_view.setFont(self.header_font)
        self.push_view.setObjectName("push_view")
        self.push_view.setText( "View Selection")
        self.push_group_layout.addWidget(self.push_view)

        #Create push_lanch button
        ##################
        self.push_launch = QtWidgets.QPushButton(self.push_group)
        self.push_launch.setFont(self.header_font)
        self.push_launch.setObjectName("push_launch")
        self.push_launch.setText( "Launch Jobs From Selection")
        self.push_group_layout.addWidget(self.push_launch)

        #Create push_quit button
        ##################
        self.push_quit = QtWidgets.QPushButton(self.push_group)
        self.push_quit.setFont(self.header_font)
        self.push_quit.setObjectName("push_quit")
        self.push_quit.setText( "Quit")
        self.push_group_layout.addWidget(self.push_quit)

        self.left_frame_layout.addWidget(self.push_group)

        #Create RegEx box
        ##################
        self.regex_group = QtWidgets.QGroupBox(self.left_frame)
        self.regex_group.setFont(self.header_font)
        self.regex_group.setTitle("RegEx Filters")
        self.regex_group_layout = QtWidgets.QFormLayout(self.regex_group)
        self.regex_group_layout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.regex_group_layout.setObjectName("regex_group_layout")

        #Create RegEx combobox and editor
        ##################
        self.regex_combobox = QtWidgets.QComboBox()
        self.regex_group_layout.addRow(QtWidgets.QLabel("Column:"), self.regex_combobox)
        self.regex_edit     = QtWidgets.QLineEdit()
        self.regex_group_layout.addRow(QtWidgets.QLabel("Regex String:"),self.regex_edit)
        self.left_frame_layout.addWidget(self.regex_group)

        #Create message box
        ##################
        self.group_message = QtWidgets.QGroupBox(self.left_frame)
        self.group_message.setFont(self.header_font)
        self.group_message.setObjectName("group_message")
        self.group_message.setTitle( "Messages")
        self.group_message_layout = QtWidgets.QVBoxLayout(self.group_message)
        self.group_message_layout.setObjectName("group_message_layout")
        self.message_box = QtWidgets.QPlainTextEdit(self.group_message)
        self.message_box.setReadOnly(True)
        self.message_box.setObjectName("message_box")
        self.group_message.setTitle( "Messages")

        #Add the message and push_button box to the left_frame
        ##################
        self.group_message_layout.addWidget(self.message_box)
        self.left_frame_layout.addWidget(self.group_message)
        self.query_tab_layout.addWidget(self.left_frame)

        #Create Right Frame and set its layout
        ##################
        self.right_frame = QtWidgets.QFrame()
        self.right_frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        # self.right_frame.setFrameShape(QtWidgets.QFrame.Panel)
        # self.right_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        # self.right_frame.setLineWidth(2)
        # self.right_frame.setMidLineWidth(0)
        # self.right_frame.setObjectName("right_frame")
        self.right_frame_layout = QtWidgets.QVBoxLayout(self.right_frame)
        self.right_frame_layout.setObjectName("right_frame_layout")

        #Create Result Box
        ##################
        self.group_results = QtWidgets.QGroupBox(self.right_frame)
        self.group_results.setFont(self.header_font)
        self.group_results.setAcceptDrops(False)
        self.group_results.setObjectName("group_results")
        self.group_results.setTitle( "Query Results")
        self.group_results_layout = QtWidgets.QVBoxLayout(self.group_results)
        # self.group_results_layout.setContentsMargins(5, 5, 5, 5)
        self.group_results_layout.setObjectName("group_results_layout")
        self.group_results.setTitle( "Query Results")
        self.table = QtWidgets.QTableView(self.group_results)
        self.table.setFrameShape(QtWidgets.QFrame.NoFrame)
        # self.table.setFrameShadow(QtWidgets.QFrame.Raised)
        self.table.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.table.setSortingEnabled(False)
        self.table.setObjectName("table")
        self.group_results_layout.addWidget(self.table)
        self.right_frame_layout.addWidget(self.group_results)

        # #Pretty Line
        # ##################
        # self.line = QtWidgets.QFrame(self.right_frame)
        # self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        # self.line.setLineWidth(3)
        # self.line.setMidLineWidth(1)
        # self.line.setFrameShape(QtWidgets.QFrame.HLine)
        # self.right_frame_layout.addWidget(self.line)

        ##################
        #Create Tab Widget
        ##################
        self.group_query = QtWidgets.QGroupBox(self.right_frame)
        self.group_query.setMaximumHeight(500)
        self.group_query.setFont(self.header_font)
        self.group_query.setFlat(True)
        self.group_query.setObjectName("group_query")
        self.group_query.setTitle( "Queries")
        self.group_query_layout = QtWidgets.QVBoxLayout(self.group_query)
        self.group_query_layout.setContentsMargins(12, 12, 12, 12)
        self.group_query_layout.setObjectName("group_query_layout")

        self.tabWidget = QtWidgets.QTabWidget(self.group_query)
        self.tabWidget.setTabPosition(QtWidgets.QTabWidget.North)
        self.tabWidget.setTabShape(QtWidgets.QTabWidget.Rounded)
        self.tabWidget.setObjectName("tabWidget")

        #Create Expert Tab
        ##################
        self.expert_tab = QtWidgets.QWidget()
        self.expert_tab.setObjectName("expert_tab")
        self.expert_tab_layout = QtWidgets.QVBoxLayout(self.expert_tab)
        self.expert_tab_layout.setObjectName("expert_tab_layout")
        #######Add Plain Text editor
        self.group_expert_editor = QtWidgets.QGroupBox(self.expert_tab)
        self.group_expert_editor.setFont(self.header_font)
        self.group_expert_editor.setObjectName("group_expert_editor")
        self.group_expert_editor.setTitle( "Direct SQL Statement (Cmd+Enter to query)")
        self.group_expert_editor_layout = QtWidgets.QHBoxLayout(self.group_expert_editor)
        self.group_expert_editor_layout.setObjectName("group_expert_editor_layout")
        self.expert_editor = QtWidgets.QPlainTextEdit(self.group_expert_editor)
        self.expert_editor.setObjectName("expert_editor")
        self.group_expert_editor_layout.addWidget(self.expert_editor)
        self.expert_tab_layout.addWidget(self.group_expert_editor)
        #######Add Query Push Button
        self.layout_push_expert = QtWidgets.QHBoxLayout()
        self.layout_push_expert.setObjectName("layout_push_expert")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.layout_push_expert.addItem(spacerItem)
        self.push_query_expert = QtWidgets.QPushButton(self.expert_tab)
        self.push_query_expert.setObjectName("push_query_expert")
        self.push_query_expert.setText("Query (Ctrl/Cmd+Enter)")
        #######Place everything in the tab widget
        self.layout_push_expert.addWidget(self.push_query_expert)
        self.expert_tab_layout.addLayout(self.layout_push_expert)
        self.tabWidget.addTab(self.expert_tab, "SQL Expert")

        #Create Form Tab
        ##################
        self.form_tab = QtWidgets.QWidget()
        self.form_tab.setObjectName("form_tab")
        self.group_form_layout = QtWidgets.QGridLayout(self.form_tab)
        self.group_form_layout.setObjectName("group_form_layout")
        self.group_form = QtWidgets.QGroupBox(self.form_tab)
        self.group_form.setFont(self.header_font)
        self.group_form.setObjectName("group_form")
        self.group_form.setTitle( "Query Form")

        #Create a scroll area for the form
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout(self.group_form)
        self.horizontalLayout_5.setContentsMargins(12, 12, 12, 12)
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.scroll_form = QtWidgets.QScrollArea(self.group_form)
        self.scroll_form.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.scroll_form.setWidgetResizable(True)
        self.scroll_form.setObjectName("scroll_form")

        self.create_scroll_area(4)

        #Add to the widget layout and add query push button
        self.horizontalLayout_5.addWidget(self.scroll_form)
        self.group_form_layout.addWidget(self.group_form, 0, 1, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.push_add_constraint = QtWidgets.QPushButton(self.form_tab)
        self.push_add_constraint.setObjectName("push_add_constraint")
        self.push_add_constraint.setText("Add Constraint")
        self.horizontalLayout.addWidget(self.push_add_constraint)
        self.push_remove_constraint = QtWidgets.QPushButton(self.form_tab)
        self.push_remove_constraint.setObjectName("push_remove_constraint")
        self.push_remove_constraint.setText("Remove Constraint")
        self.horizontalLayout.addWidget(self.push_remove_constraint)
        self.push_query = QtWidgets.QPushButton(self.form_tab)
        self.push_query.setObjectName("push_query")
        self.push_query.setText( "Query (Ctrl/Cmd+Enter)")
        self.horizontalLayout.addWidget(self.push_query)
        self.group_form_layout.addLayout(self.horizontalLayout, 1, 1, 1, 1)

        #Add explanation of form
        self.plainTextEdit = QtWidgets.QPlainTextEdit(self.form_tab)
        self.plainTextEdit.setAutoFillBackground(False)
        self.plainTextEdit.setReadOnly(True)
        self.plainTextEdit.setObjectName("plainTextEdit")
        self.plainTextEdit.setPlainText( "The default table for the form is a join of job, relax_job, finaltraj and struct, as well as a left join with the surface, molecule, and bulk tables. If you would like a different join please use the expert mode.\n")
        self.group_form_layout.addWidget(self.plainTextEdit, 0, 0, 2, 1)
        self.tabWidget.addTab(self.form_tab, "")
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.form_tab),"SQL Form")
        self.group_query_layout.addWidget(self.tabWidget)
        self.right_frame_layout.addWidget(self.group_query)
        self.query_tab_layout.addWidget(self.right_frame)
        self.right_frame_layout.setStretch(1,1)
        self.query_tab_layout.setStretch(1,1)

        self.central_tabWidget.addTab(self.query_tab,"Query GUI")
        # self.setCentralWidget(self.centralwidget)
        QtCore.QMetaObject.connectSlotsByName(self)



    def create_scroll_area(self, n_constraints : int = 3) -> None:
        """
        Creates the scroll area for the SQL form
        Mainly used for the add and remove constraints buttons

        Parameters
        ----------
        n_constraints : int
            number of constraints visible in the ui of the where clause
        """
        #Place the form in the contents of the scroll area
        self.scroll_contents = QtWidgets.QWidget()
        self.scroll_contents.setGeometry(QtCore.QRect(0, 0, 525, 330))
        self.scroll_contents.setObjectName("scroll_contents")
        self.scroll_contents_layout = QtWidgets.QFormLayout(self.scroll_contents)
        self.scroll_contents_layout.setLabelAlignment(QtCore.Qt.AlignCenter)
        self.scroll_contents_layout.setObjectName("scroll_contents_layout")

        #Create Select label
        self.label_select = QtWidgets.QLabel(self.scroll_contents)
        self.label_select.setFont(self.header_font)
        self.label_select.setAlignment(QtCore.Qt.AlignCenter)
        self.label_select.setObjectName("label_select")
        self.label_select.setText( "SELECT")
        self.scroll_contents_layout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_select)

        #Create Select edit box
        self.select_edit = QtWidgets.QLineEdit(self.scroll_contents)
        self.select_edit.setObjectName("select_edit")
        self.select_edit.setText( "*")
        self.scroll_contents_layout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.select_edit)

        #Create Where label
        self.label_where = QtWidgets.QLabel(self.scroll_contents)
        self.label_where.setFont(self.header_font)
        self.label_where.setAlignment(QtCore.Qt.AlignCenter)
        self.label_where.setObjectName("label_where")
        self.label_where.setText( "WHERE")
        self.scroll_contents_layout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_where)

        #Create Constraint fields
        self.n_constraints = 0
        self.constraint_columns    = [] # type: ignore
        self.constraint_conditions = [] # type: ignore
        self.constraint_edits      = [] # type: ignore
        self.layout_constraints    = [] # type: ignore

        default_columns = ['job_id','user','system_type']+['']*(n_constraints-3)
        [self.create_constraint(col) for col in default_columns] # type: ignore

        #Create Group By label
        self.label_group_by = QtWidgets.QLabel(self.scroll_contents)
        self.label_group_by.setFont(self.header_font)
        self.label_group_by.setAlignment(QtCore.Qt.AlignCenter)
        self.label_group_by.setObjectName("label_group_by")
        self.label_group_by.setText( "GROUP BY")
        self.scroll_contents_layout.setWidget(self.scroll_contents_layout.count(), QtWidgets.QFormLayout.LabelRole, self.label_group_by)
        self.group_by_edit = QtWidgets.QLineEdit(self.scroll_contents)
        self.group_by_edit.setObjectName("group_by_edit")
        self.scroll_contents_layout.setWidget(self.scroll_contents_layout.count(), QtWidgets.QFormLayout.FieldRole, self.group_by_edit)

        #Create Limit label
        self.label_limit = QtWidgets.QLabel(self.scroll_contents)
        self.label_limit.setFont(self.header_font)
        self.label_limit.setAlignment(QtCore.Qt.AlignCenter)
        self.label_limit.setObjectName("label_limit")
        self.label_limit.setText( "LIMIT")
        self.scroll_contents_layout.setWidget(self.scroll_contents_layout.count(), QtWidgets.QFormLayout.LabelRole, self.label_limit)
        self.limit_edit = QtWidgets.QLineEdit(self.scroll_contents)
        self.limit_edit.setObjectName("limit_edit")
        self.scroll_contents_layout.setWidget(self.scroll_contents_layout.count(), QtWidgets.QFormLayout.FieldRole, self.limit_edit)
        self.scroll_form.setWidget(self.scroll_contents)

    def create_constraint(self, column_name : str = '') -> None:
        """
        A UI function for making each row for a constraint in the SQL Form.

        Parameters
        ----------
        column_name : str
            name of column to be set in the line edit box of the constraint row
        """
        #Create Column Editor
        self.constraint_columns.append(QtWidgets.QLineEdit(self.scroll_contents))
        self.constraint_columns[self.n_constraints].setAlignment(QtCore.Qt.AlignCenter)
        self.constraint_columns[self.n_constraints].setText(column_name)
        self.constraint_columns[self.n_constraints].setObjectName("constraint_column[self.n_constraints]")
        self.scroll_contents_layout.setWidget(self.n_constraints+3, QtWidgets.QFormLayout.LabelRole, self.constraint_columns[self.n_constraints])

        self.layout_constraints.append(QtWidgets.QHBoxLayout())
        self.layout_constraints[self.n_constraints].setObjectName("layout_constraint_1")

        #Create Condition combo box
        self.constraint_conditions.append(QtWidgets.QComboBox(self.scroll_contents))
        self.constraint_conditions[self.n_constraints].setObjectName("constraint_conditions_{}".format(self.n_constraints))
        combo_box_strings = ["Equal","Not Equal","Like","Not Like","In"]
        [self.constraint_conditions[self.n_constraints].addItem(item) for item in combo_box_strings]
        self.layout_constraints[self.n_constraints].addWidget(self.constraint_conditions[self.n_constraints])

        #Create Constraint Editor
        self.constraint_edits.append(QtWidgets.QLineEdit(self.scroll_contents))
        self.constraint_edits[self.n_constraints].setObjectName("constraint_edits[self.n_constraints]")
        self.layout_constraints[self.n_constraints].addWidget(self.constraint_edits[self.n_constraints])
        self.scroll_contents_layout.setLayout(self.n_constraints+3, QtWidgets.QFormLayout.FieldRole, self.layout_constraints[self.n_constraints])
        self.n_constraints += 1

    @QtCore.pyqtSlot(int)                                            # type: ignore
    def on_view_horizontalHeader_sectionClicked(self, logicalIndex): # type: ignore
        self.logicalIndex   = logicalIndex
        self.menuValues     = QtWidgets.QMenu(self)
        self.signalMapper   = QtCore.QSignalMapper(self)

        self.regex_combobox.blockSignals(True)
        self.regex_combobox.setCurrentIndex(self.logicalIndex)
        self.regex_combobox.blockSignals(False)

        valuesUnique = [    str(self.model.record(row).value(self.logicalIndex))
                            for row in range(self.model.rowCount())
                            ]
        if len(set(valuesUnique)) == self.model.rowCount():
            return
        actionAll = QtWidgets.QAction("All", self)
        actionAll.triggered.connect(self.on_actionAll_triggered)
        self.menuValues.addAction(actionAll)
        self.menuValues.addSeparator()

        for actionNumber, actionName in enumerate(sorted(list(set(valuesUnique)))):
            action = QtWidgets.QAction(actionName, self)
            self.signalMapper.setMapping(action, actionNumber)
            action.triggered.connect(self.signalMapper.map)
            self.menuValues.addAction(action)

        self.signalMapper.mapped.connect(self.on_signalMapper_mapped)

        headerPos = self.table.mapToGlobal(self.horizontalHeader.pos())

        posY = headerPos.y() + self.horizontalHeader.height()
        posX = headerPos.x() + self.horizontalHeader.sectionPosition(self.logicalIndex)

        self.menuValues.exec_(QtCore.QPoint(posX, posY))

    @QtCore.pyqtSlot()                                         # type: ignore
    def on_actionAll_triggered(self)->None:                    # type: ignore
        filterColumn = self.logicalIndex
        filterString = QtCore.QRegExp( "",
                                        QtCore.Qt.CaseInsensitive,
                                        QtCore.QRegExp.RegExp
                                        )
        self.proxy.setFilterRegExp(filterString)
        self.proxy.setFilterKeyColumn(filterColumn)

    @QtCore.pyqtSlot(int)                                      # type: ignore
    def on_signalMapper_mapped(self, i : typ.Any)->None:       # type: ignore
        stringAction = self.signalMapper.mapping(i).text()
        filterColumn = self.logicalIndex
        filterString = QtCore.QRegExp( stringAction,
                                        QtCore.Qt.CaseSensitive,
                                        QtCore.QRegExp.FixedString
                                        )
        self.proxy.setFilterRegExp(filterString)
        self.proxy.setFilterKeyColumn(filterColumn)

    @QtCore.pyqtSlot(str)                                      # type: ignore
    def on_lineEdit_textChanged(self, text : typ.Any) -> None: # type: ignore
        search = QtCore.QRegExp( text,
                                    QtCore.Qt.CaseInsensitive,
                                    QtCore.QRegExp.RegExp
                                    )

        self.proxy.setFilterRegExp(search)

    @QtCore.pyqtSlot(int)
    def on_comboBox_currentIndexChanged(self, index):          # type: ignore
        self.proxy.setFilterKeyColumn(index)

if __name__ == "__main__":
    import sys,os
    os.environ['QT_API']='pyqt5'
    sys.argv.append('--disable-web-security')
    app = QtWidgets.QApplication(sys.argv)
    app.setStyleSheet(qdarkstyle.load_stylesheet_from_environment(is_pyqtgraph=False))

    gui = JobMakerGUI()
    gui.showMaximized()
    sys.exit(app.exec_())



# typing Modules
import typing as typ

# External Modules

# Internal Modules
from catalog.jobs.jobs        import Job
from catalog.jobs.versions    import current_version
from catalog.datalog.db_utils import Query
# from catalog.misc.sql_utils   import STRJOB_EXACT_,STRJOB_GRAPH_,job,JOB
################################################################################
class RepeatChecker(object):
    """
    Interface for checking repeat jobs
    Preserve this object throughout submitting jobs in one session
    to avoid constantly scraping fireworks database
    """

    def __init__(self
                ,granularity : str = 'exact'
                ) -> None:

        assert granularity in ['graph','exact'], 'Invalid option for "granularity"'
        self.granularity=granularity
        self._scrape_incomplete_jobs()

    def add_incomplete_job(self
                          ,job : Job
                          ) -> None:
        """
        Add a newly created job to the list for repeat-checking
        """
        self.incomplete_jobs.append(job.to_string(self.granularity))

    def check_repeat(self
                    ,job : Job
                    ) -> bool:
        """
        Returns True if job is a repeat
        """
        #HACK FOR GETTING AROUND BROKEN CHECKER
        return False

        job_str = job.to_string(self.granularity)
        return (  self._repeat_incomplete(job_str)
               or self._repeat_complete(job_str))

    def _repeat_incomplete(self
                          ,job_str : str
                          ) -> bool:
        """
        Check current list of incomplete jobs for a repeat
        """
        return job_str in self.incomplete_jobs

    def _repeat_complete(self
                        ,job_str : str
                        ) -> bool:
        """Check CataLog database for a particular job"""
        constraint = STRJOB_EXACT_ if self.granularity=='exact' else STRJOB_GRAPH_
        constraints = [constraint(job_str)] # type: ignore
        return Query(constraints=constraints,table=job
                       ,db_path=self.db_path).any_query(JOB)

    def _scrape_incomplete_jobs(self) -> None:
        """
        Initialize incomplete job list from FireWorks database
        """
        from catalog.fw.incomplete import incompleteFWIDS,lpad
        incomplete_fwids = incompleteFWIDS()
        params = [lpad.get_fw_dict_by_id(i)['spec']['params'] for i in incomplete_fwids]
        self.incomplete_jobs = [current_version.process_dict(p).to_string(self.granularity) for p in params] # type: typ.List[str]



# Typing modules
import typing as typ

#External Modules
import json,math,itertools
import numpy as np                                                      # type: ignore
from copy import deepcopy

#--ase Imports
import ase                                                              # type: ignore
from ase.io import read,write                                           # type: ignore
from ase.constraints import FixAtoms                                    # type: ignore
from ase.visualize import view                                          # type: ignore
from ase.gui.images import Images                                       # type: ignore
from ase.gui.gui import GUI                                             # type: ignore

#--pymatgen Imports
import pymatgen as pmg 		                                            # type: ignore
from pymatgen.io.ase 		import AseAtomsAdaptor                      # type: ignore
from pymatgen.core.surface 	import SlabGenerator,Slab                   # type: ignore
from pymatgen.analysis.adsorption import AdsorbateSiteFinder,reorient_z # type: ignore

# Internal Modules
import catalog.data.data_solids_wPBE as soldata
import catalog.data.data_sol53_54_57_58_bm32_se30 as surfdata
from .utilities import flatten
from .print_parse import roundfloat


"""
Functions for interacting with ASE


Keld Data related
    get_keld_data
    get_expt_surf_energy

Element Lists
    nonmetal_symbs
    nonmetals
    mag_elems
    mag_nums
    paw_elems
    symbols2electrons

Atoms object manipulation
    traj_to_json
    json_to_traj
    classify_system
    make_atoms
    cell_to_param
    restore_magmom

Geometry
    angle
    dihedralz

Site Related
    get_sites


"""
################################################################################

# Keld Data Related
#-----------------
def get_keld_data(name  : str
                 ,k 	: str
                 ) -> typ.Any:
    """
    Docstring
    """

    name = name.split('_')[0] #sometimes suffix ('bulkmod') is appended to name
    try: return soldata.data[soldata.getKey[name] ][k]
    except KeyError:
        try: return soldata.data[soldata.getKey[name] ][k+' kittel'] #for 'bulkmodulus' / 'bulkmodulus kittel'
        except KeyError: return None

###########################
# Atoms object manipulation
###########################

def traj_to_json(atoms : ase.Atoms) -> str:
    """
    Serialize an Atoms object in a human readable way
    """

    atomdata = []
    const    = atoms.constraints
    if const: const_list = const[0].get_indices().tolist()
    else:     const_list = []

    atoms.wrap()
    for a in atoms: atomdata.append({'number'		: int(a.number)
                                    ,'x'			: roundfloat(a.x)
                                    ,'y'			: roundfloat(a.y)
                                    ,'z'			: roundfloat(a.z)
                                    ,'magmom'		: roundfloat(a.magmom)
                                    ,'tag'			: int(a.tag)
                                    ,'constrained'	: int(a.index in const_list)
                                    ,'index'		: int(a.index)})

    out = {'cell': [[roundfloat(x) for x in xx] for xx in atoms.get_cell().tolist()]
          ,'atomdata':atomdata}

    return json.dumps(out)




def json_to_traj(raw_json:str)->ase.Atoms:
    """
    Inverse of traj_to_json
    """

    raw_atoms = json.loads(raw_json)
    try:
        atom_data = raw_atoms['atom_data']
    except KeyError:
        atom_data = raw_atoms['atomdata']

    pos = np.array([[a[q] for a in atom_data] for q in ['x','y','z']]).T

    fix = FixAtoms([a['index'] for a in atom_data if a['constrained']])

    atoms = ase.Atoms(numbers     = [a['number'] for a in atom_data]
                     ,cell        = raw_atoms['cell']
                     ,positions   = pos
                     ,magmoms     = [a['magmom'] for a in atom_data]
                     ,tags        = [a['tag'] for a in atom_data]
                     ,constraint  = fix)
    return atoms


def cell_to_param(cell : np.array) -> typ.Tuple[float,float,float,float,float,float]:
    """
    ANGLES ARE IN RADIANS
    """
    a = np.linalg.norm(cell[0])
    b = np.linalg.norm(cell[1])
    c = np.linalg.norm(cell[2])
    alpha = angle(cell[1],cell[2])
    beta  = angle(cell[0],cell[2])
    gamma = angle(cell[0],cell[1])
    return (a,b,c,alpha,beta,gamma)

def classify_system(atoms : ase.Atoms) -> str:
    cutoff    = 6  # A
    atoms.center() # In case dealing with Nerds Rope & Co.
    minx,miny,minz,maxx,maxy,maxz = 1000,1000,1000,-1000,-1000,-1000
    for a in atoms:
        if a.position[0] < minx: minx =  a.position[0]
        if a.position[1] < miny: miny =  a.position[1]
        if a.position[2] < minz: minz =  a.position[2]
        if a.position[0] > maxx: maxx =  a.position[0]
        if a.position[1] > maxy: maxy =  a.position[1]
        if a.position[2] > maxz: maxz =  a.position[2]

    cell_abc  = list(map(np.linalg.norm,atoms.get_cell()[:3]))
    thickness = [maxx-minx,maxy-miny,maxz-minz]
    pbc       = [cell_abc[i]-thickness[i] < cutoff for i in [0,1,2]]

    if   all(pbc): return 'bulk'
    elif any(pbc): return 'surface'
    else:          return 'molecule'

def make_atoms(*args : str) -> ase.Atoms:
    n,p_t,c,m,cs = list(map(json.loads,args))
    return ase.Atoms(numbers=n,positions=list(zip(*p_t))
            ,cell=c,magmoms=m
            ,constraint=FixAtoms(np.nonzero(cs)[0]))

def remove_atom(atoms_obj       : ase.Atoms
               ,index_to_delete : int
               )->ase.Atoms:
    atoms_obj_watom_removed = atoms_obj.copy()
    fixed_inds              = get_list_of_fixed_indices(atoms_obj)
    fixed_inds              = [x if x<=index_to_delete else x-1 for x in fixed_inds]
    del atoms_obj_watom_removed[:]
    for ind,atom_obj in enumerate(atoms_obj):
        if not ind == index_to_delete:
            atoms_obj_watom_removed.append(atom_obj)
    atoms_obj_watom_removed.set_constraint(FixAtoms(indices=fixed_inds))
    return atoms_obj_watom_removed

def remove_atoms(atoms_obj         : ase.Atoms
                ,indices_to_delete : typ.List[int]
                )-> ase.Atoms      :
    atoms_obj_watoms_removed     = atoms_obj.copy()
    for ind in sorted(indices_to_delete,reverse = True):
        atoms_obj_watoms_removed = remove_atom(atoms_obj_watoms_removed,ind)
    return atoms_obj_watoms_removed

def get_list_of_fixed_indices(atoms_obj : ase.Atoms)->list:
    constraints          = atoms_obj.constraints
    fixed_inds           = [] # type: list
    for constraint in constraints:
        if isinstance(constraint,FixAtoms):
            fixed_inds += list(constraint.get_indices())
    return fixed_inds

def fix_negative_unit_cells(atoms_obj : ase.Atoms)-> ase.Atoms:
    cell = atoms_obj.cell
    signs = np.sign(np.sum(cell,axis = 1))
    for i, sign in enumerate(signs):
        cell[i] *= sign
    new_atoms_obj = atoms_obj.copy()
    new_atoms_obj.set_cell(cell)
    new_atoms_obj.wrap(pbc = [True,True,True])
    return new_atoms_obj

##########
# Geometry
##########
def angle(v1 : np.array
         ,v2 : np.array
         ) -> float:
    return np.arccos(np.dot(v1,np.transpose(v2))/(np.linalg.norm(v1)*np.linalg.norm(v2)))
def dihedral(v1 : np.array
            ,v2 : np.array
            ,v3 : np.array
            ) -> float:
    """
    http://azevedolab.net/resources/dihedral_angle.pdf
    """
    x12 = np.cross(v1,v2)
    x23 = np.cross(v2,v3)
    n1  = x12/np.linalg.norm(x12)
    n2  = x23/np.linalg.norm(x23)
    u1  = n2
    u3  = v2/np.linalg.norm(v2)
    u2  = np.cross(u3,u1)
    cos_theta = np.dot(n1,u1)
    sin_theta = np.dot(n1,u2)
    theta = -math.atan2(sin_theta,cos_theta)
    return theta


##############
# Element lists
###############
nonmetal_symbs   = ['H','C','N','P','O','S','Se','F','Cl','Br','I','At','He','Ne','Ar','Kr','Xe','Rn']
nonmetals        = [ase.data.chemical_symbols.index(x) for x in nonmetal_symbs]

relevant_atoms = [1,3,4,6,7,8,9,11,12,13,14,16,17] \
                    + list(range(19,36))+list(range(37,52))+[55,56]+list(range(72,80))

relevant_nonmetals = list(set(nonmetals) & set(relevant_atoms))

mag_elems        = ['Fe','Mn','Cr','Co','Ni']
mag_nums         = [24,25,26,27,28]
paw_elems        = ['Li','Be','Na','Mg','K','Ca','Rb','Sr','Cs','Ba','Zn']

def symbols2electrons(symbols : typ.List[str]
                     ,psp 	  : str ='gbrv'
                     ) -> int : # added Be, Cd on my own
    symbdict = {'gbrv':{'Ag':19,'Al':3,'As':5,'Au':11,'Ba':10,'Br':7,'B':3,'Be':4,'Ca':10,'Cd':12,'Co':17,'Cr':14,'Cs':9,'C':4,'Cu':19,'Fe':16,'F':7,'Ga':19,'Ge':14,'Hf':12,'Hg':12,'H':1,'In':13,'Ir':15,'I':7,'K':9,'La':11,'Li':3,'Mg':10,'Mn':15,'Mo':14,'Na':9,'Nb':13,'Ni':18,'N':5,'Os':16,'O':6,'Pb':14,'Pd':16,'Pt':16,'P':5,'Rb':9,'Re':15,'Rh':15,'Ru':16,'Sb':15,'Sc':11,'Se':6,'Si':4,'Sn':14,'Sr':10,'S':6,'Ta':13,'Tc':15,'Te':6,'Ti':2,'Tl':13,'V':13,'W':14,'Y':11,'Zn':20,'Zr':12}}
    return sum([symbdict[psp][x] for x in symbols])

#############
# Site Related
###############
def make_pmg_slab(a : ase.Atoms,facet : typ.List[int]) -> pmg.Structure:
    species             = a.get_chemical_symbols()
    coords              = a.get_positions()
    miller_index        = facet
    oriented_unit_cell  = AseAtomsAdaptor.get_structure(a)
    shift               = 0    #???????
    scale_factor        = None #???????
    return Slab(oriented_unit_cell.lattice, species, coords, miller_index,oriented_unit_cell, shift, scale_factor,coords_are_cartesian=True)

def reorient_z_ase(a : ase.Atoms,facet : typ.List[int]) -> ase.Atoms:
    slab                     = make_pmg_slab(a,facet)
    slab                     = reorient_z(slab)
    out_atoms_obj            = AseAtomsAdaptor.get_atoms(slab)
    out_atoms_obj.cell[2,:2] = 0.
    return out_atoms_obj

def get_sites(a 		  : ase.Atoms
             ,facet 	  : typ.List[int]
             ,site_type   : str  		 	= 'all'
             ,symm_reduce : float 			= 0.0
             ,height 	  : float 			= 0.9
             ) -> typ.Any:
    assert site_type in ['all','bridge','ontop','hollow'], 'Please supply a valid site_type'

    slab  = make_pmg_slab(a,facet)
    slab  = reorient_z(slab)
    sites = AdsorbateSiteFinder(slab,height = height).find_adsorption_sites(symm_reduce=symm_reduce, distance = 0)[site_type]
    return sites



def get_mic_distance(p1       : np.array
                    ,p2 	  : np.array
                    ,cell     : np.array
                    ,pbc 	  : typ.List[int]
                    ,dis_ind  : str = 'xyz'
                    ) -> float:
    """ This method calculates the shortest distance between p1 and p2
         through the cell boundaries defined by cell and pbc.
         This method works for reasonable unit cells, but not for extremely
         elongated ones.
    """
    ct = cell.T
    pos = np.mat((p1, p2))
    scaled = np.linalg.solve(ct, pos.T).T
    for i in range(3):
        if pbc[i]:
            scaled[:, i] %= 1.0
            scaled[:, i] %= 1.0
    P = np.dot(scaled, cell)

    pbc_directions = [[-1, 1] * int(direction) + [0] for direction in pbc]
    translations = np.mat(list(itertools.product(*pbc_directions))).T
    p0r = np.tile(np.reshape(P[0, :], (3, 1)), (1, translations.shape[1]))
    p1r = np.tile(np.reshape(P[1, :], (3, 1)), (1, translations.shape[1]))
    dp_vec = p0r + ct * translations
    if dis_ind == 'xyz':
        squared_dis = np.power((p1r - dp_vec), 2).sum(axis=0)
    elif dis_ind =='xy':
        squared_dis = np.power((p1r - dp_vec)[0:2], 2).sum(axis=0)
    else:
        raise ValueError('Please provide valid direction to include in distance \'xy\' or \'xyz\'')
    d = np.min(squared_dis)**0.5
    return d

def collision(position           : np.array
             ,atom               : ase.Atom
             ,cell               : np.array
             ,pbc                : typ.List[int]
             ,dis_ind            : str   = 'xyz'
             ,collision_distance : float = 2.
             )->bool:
    distance = get_mic_distance(position,atom.position,cell,pbc)
    print (distance)
    return distance < (ase.data.covalent_radii[atom.number])*2/1.124 or distance<collision_distance

def get_mic_vector(p1      : np.array
                  ,p2 	   : np.array
                  ,cell    : np.array
                  ,pbc 	   : typ.List[int]
                  ,dis_ind : str = 'xyz'
                  )-> np.array:
    """ This method calculates the shortest distance between p1 and p2
         through the cell boundaries defined by cell and pbc.
         This method works for reasonable unit cells, but not for extremely
         elongated ones.
    """
    ct = cell.T
    pos = np.mat((p1, p2))
    scaled = np.linalg.solve(ct, pos.T).T
    for i in range(3):
        if pbc[i]:
            scaled[:, i] %= 1.0
            scaled[:, i] %= 1.0
    P = np.dot(scaled, cell)
    pbc_directions = [[-1, 1] * int(direction) + [0] for direction in pbc]
    translations = np.mat(list(itertools.product(*pbc_directions))).T
    p0r = np.tile(np.reshape(P[0, :], (3, 1)), (1, translations.shape[1]))
    p1r = np.tile(np.reshape(P[1, :], (3, 1)), (1, translations.shape[1]))
    dp_vec = p0r + ct * translations
    if dis_ind == 'xyz':
        squared_dis = np.power((p1r - dp_vec), 2).sum(axis=0)
    elif dis_ind =='xy':
        squared_dis = np.power((p1r - dp_vec)[0:2], 2).sum(axis=0)
    else:
        raise ValueError('Please provide valid direction to include in distance \'xy\' or \'xyz\'')
    d = np.min(squared_dis)**0.5
    return d


def delete_atom(atoms_obj : ase.Atoms
               ,ind : int
               ) -> ase.Atoms:
    """
    Returns a new Atoms object with an atom removed at the index
    """
    new_atoms_obj = atoms_obj.copy()
    del new_atoms_obj[ind]
    return new_atoms_obj

def delete_vacancy(parent_atoms_obj : ase.Atoms
                  ,child_atoms_obj  : ase.Atoms
                  ) -> ase.Atoms:
    """
    Docstring
    """
    vac_pos, vac_ind = get_vacancy_pos(parent_atoms_obj, child_atoms_obj)
    return delete_atom(parent_atoms_obj,vac_ind)

def delete_adsorbates(atoms_obj 		: ase.Atoms
                     ,adsorbates_list   : typ.List[typ.List[int]]
                     ) -> ase.Atoms:
    """
    Helpful docstring
    """
    new_atoms_obj = atoms_obj.copy()
    reverse_ordered_adsorbate_list = np.sort(flatten(adsorbates_list))[::-1]
    for ind in reverse_ordered_adsorbate_list:
        new_atoms_obj = delete_atom(new_atoms_obj, ind)
    return new_atoms_obj

def did_it_restructure(atoms_1 				: ase.Atoms
                      ,atoms_2				: ase.Atoms
                      ,restructure_criteria : int 		 = 1
                      ) -> bool:
    """
    Docstring explaining how this algorithm works
    """
    distances_1 = np.linalg.norm(atoms_1.positions, axis = 1)
    distances_2 = np.linalg.norm(atoms_2.positions, axis = 1)
    max_dis = np.max(np.abs(distances_2-distances_1))
    vert_dis = np.max(atoms_1.positions[:,2] - atoms_2.positions[:,2])
    return np.max(np.abs(distances_2-distances_1))>restructure_criteria

def get_vacancy_pos(parent : ase.Atoms
                   ,child  : ase.Atoms
                   ) -> typ.Tuple[np.array,int]:
    """
    Returns the list of length three indicating the location of the vacancy
    position on the slab using the bare_slab parent indicated in the database.
    The parent should be the atoms object of the relaxed bare slab with no
    """
    if len(parent) < len(child): raise ValueError("can't get a vacancy when len(parent)=%d and len(child)=%d"%(len(parent) , len(child)))
    #Find the distances between each atom in the parent object and each of the atoms in the child object
    #Returns list with a length equal to that of the parents object
    distances = map(lambda pos: map(np.linalg.norm,(pos-child.positions)), parent.positions)
    #Find the nearest neighbor between the parent atom and the child atoms
    min_dis = map(np.min, distances)
    #the atom with the largest distance between the parent and child atoms is assumed to be
    # at the location of the vacancy
    return parent.positions[np.argmax(min_dis)], np.argmax(min_dis)

def restore_magmom(trajjson : str) -> str:
    """
    docstring
    """
    traj = json_to_traj(trajjson)
    try:
        mags = traj.get_magmoms()
        if any([x>0 for x in mags]): traj.set_initial_magnetic_moments([3 if e in mag_elems else 0 for e in traj.get_chemical_symbols()])
    except:
        pass
    return traj_to_json(traj)


def smart_view(traj_file : str
              ,expr      : str   = ''
              ) -> None:
    images = Images(images = read(traj_file,':'))
    gui = GUI(images = images, expr= expr)
    gui.run()
    return gui.images[-1]

import os
import subprocess
import sys
import tempfile


def new_view(atoms  : ase.Atoms
            ,data   : typ.Any                = None
            ,viewer : str                    = 'ase'
            ,graph  : str                    = ""
            ,repeat : typ.Tuple[int,int,int] = None
            ,block  : bool                   = False):
    # Ignore for parallel calculations:

    vwr = viewer.lower()

    if vwr == 'ase':
        format = 'traj'
        command = sys.executable + ' -m ase gui'
        if repeat is not None:
            command += ' --repeat={},{},{}'.format(*repeat)
            repeat = None
        command += ' --graph={}'.format(graph)


    fd, filename = tempfile.mkstemp('.' + format, 'ase-')
    if repeat is not None:
        atoms = atoms.repeat()
    if data is None:
        write(filename, atoms, format=format)
    else:
        write(filename, atoms, format=format, data=data)
    if block:
        subprocess.call(command.split() + [filename])
        os.remove(filename)
    else:
        subprocess.Popen(command.split() + [filename])
        subprocess.Popen(['sleep 60; rm {0}'.format(filename)], shell=True)



###################
#Archived Scripts
###################
#
# def get_expt_surf_energy(name : str) -> typ.Optional[float]:
#     if 'x' not in name or '_' not in name: return None
#
#     metal,crystal = name.split('_')[0].split('-')
#     facet = name.split('_')[1].replace(',','')
#
#     if crystal=='hcp' and facet == '001':
#         facet = '0001'
#     try:
#         #convert J/m^2 to eV/A^2
#         return surfdata.get_exp_surface_energy(metal+facet)[1]*0.0624 # type: ignore
#     except KeyError:
#         return None
# def show_sites(a 	 	   : ase.Atoms
#               ,facet 	   : typ.List[int]
#               ,site_type   : str   = 'all'
#               ,symm_reduce : float = 0.01
#               ,height 	   : float = 1.0
#               )-> typ.Any:
#     slab = make_pmg_slab(a,facet)
#     plot_slab(slab,plt.gca(),repeat=3,site_type = site_type, symm_reduce=symm_reduce, height = height)
#     plt.show() # user looks, closes plot to continue



#Typing imports
import typing as typ

#External imports

#Internal Imports
import parser # type: ignore

class CLICommand:
    short_description = "ASE's graphical user interface"
    description = ('ASE-GUI.  See the online manual '
                   '(https://wiki.fysik.dtu.dk/ase/ase/gui/gui.html) '
                   'for more information.')

    @staticmethod
    def add_arguments(parser : 'parser') -> None: # type: ignore
        pass

    @staticmethod
    def run(args):
        from catalog.gui.main import main
        main()
